
/*
 * $Id: SrvXact.cc,v 1.1.1.3 1999/02/06 03:20:57 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include <strstream.h>

#include "Clock.h"
#include "Rnd.h"
#include "Connection.h"
#include "StatPing.h"
#include "StatPhaseMgr.h"
#include "SharedOpts.h"
#include "SrvOpts.h"
#include "Server.h"
#include "SrvXact.h"
#include "globals.h"


void SrvXact::reset() {
	Xaction::reset();
	theOwner = 0;
	theRepSize = -1;
	theProducedSize = 0;
	theReqHdr.reset();
}

void SrvXact::exec(Server *anOwner, Connection *aConn, Time delay) {
	Assert(anOwner);
	theOwner = anOwner;
	theConn = aConn; // XXX: kludge to save connection till doStart
	if (delay > 0)
		sleepFor(delay);
	else
		doStart();
}

void SrvXact::doStart() {
	start(theConn);

	theState = stHdrWaiting;
	if (!theConn->theRd.theReserv)
		theConn->theRd.start(this);
}

// XXX: should accept errno
void SrvXact::finish(bool ok) {
	stop(ok ? 0 : -1); // errno
	theOwner->noteXactDone(this);
}

void SrvXact::stopStats() {
	TheStatPhaseMgr->xactRdSz(theProducedSize, theLifeTime);
	Xaction::stopStats();
}

void SrvXact::wakeUp(const Alarm &a) {
	AlarmUser::wakeUp(a);
	doStart();
}

void SrvXact::noteReadReady(int) {
	Assert(theState == stHdrWaiting);

	theConn->read();
	if (theConn->bad()) {
		finish(false);
		return;
	}

	if (theReqHdr.parse(theConn->theRdBuf.content(), theConn->theRdBuf.contSize())) {
		theConn->theRd.stop(this);

		// dump request headers
		if (TheOpts.doPrintReqs)
			cout.write(theConn->theRdBuf.content(), theReqHdr.theHdrSize);

		if (theReqHdr.theObjId <= 0 || !theReqHdr.theXactId) {
			finish(false); // foreign request
			return;
		}
		makeRep(theConn->theWrBuf);
		Assert(theRepSize > 0);
		if (!theConn->theWr.theReserv)
			theConn->theWr.start(this);
		theState = stSpaceWaiting;
		return;
	}

	if (theConn->theRdBuf.full()) { // too big header
		finish(false);
		return;
	}

	if (theConn->exhausted()) { // premature eof
		finish(false);
		return;
	}

	if (!theConn->theRd.theReserv) // need read more
		theConn->theRd.start(this);
}

void SrvXact::noteWriteReady(int) {
	Assert(theState == stSpaceWaiting);
	WrBuf &buf = theConn->theWrBuf;
	Assert(theProducedSize > 0 || buf.contSize() > 0);
	
	// keep the write buffer full
	buf.pack();
	buf.appended(Min(buf.spaceSize(),
		(Size)(theRepSize - theProducedSize - buf.contSize())));

	//cerr << here << "noteWriteReady: " << theProducedSize << " < " << theRepSize << endl;

	const Size sz = theConn->write();
	if (theConn->bad()) {
		finish(false);
		return;
	}

	Assert(sz >= 0);
	theProducedSize += sz;
	if (theProducedSize >= theRepSize) { // done
		Assert(theProducedSize == theRepSize);
		finish(true);
		return;
	}

	if (!theConn->theWr.theReserv) // need write more
		theConn->theWr.start(this);
}

// always reply with 200 OK
void SrvXact::makeRep(WrBuf &buf) {

	const Size clen = calcRepSize();
	const char *repStart = buf.space();

	// XXX: strstream is not the most efficient implementation
	// XXX: can be further optimized using const Strings
	ostrstream os(buf.space(), buf.spaceSize());

	os
		<< "HTTP/1.0 200 OK" << crlf
		<< "Content-Length: " << clen << crlf;

	if (TheSrvOpts.isRepCachable)
		os << "Cache-Control: public,max-age=31536000" << crlf;
	else
		os << "Cache-Control: private,no-cache" << crlf;

	if (theReqHdr.theXactId)
		os << "X-Xact: " << theReqHdr.theXactId.genMutant() << crlf;

	os
		<< "Connection: close" << crlf
		<< crlf;

	buf.appended(os.tellp());

	const Size hsize = buf.space() - repStart; // header size
	theRepSize = clen + hsize;

	// dump reply headers
	if (TheOpts.doPrintReps)
		cout.write(repStart, hsize);
}

Size SrvXact::calcRepSize() const {
	// XXX: hard coded exponential distribution
	static ExpDistr TheSizeDistr(new RndGen(), TheSrvOpts.theRepSize);
	const int objId = theReqHdr.theObjId;

	// seed random generator with objId
	TheSizeDistr.rndGen().seed(objId);
	const Size sz = (size_t) TheSizeDistr();
	Assert(sz >= 0);

	// XXX: hard coded size limits:
	// most popular objects should not be too big (100*id + 10KB)
	// objects should not exceed 50MB
	static Size maxSize = Size::MB(50);
	return sz % Min((Size)(100*objId + Size::KB(10)), maxSize);
}
